//	input.c
//
// routines called from midifile.c to process data extracted from
// the Midi File.
//

#include "mtm.h"
#include <errno.h>

static int format;		// Midifile format 0,1[,2?]
static Eventp reuse;
static char *nnam[] = { "C", "C#", "D", "D#", "E", "F", "F#", "G",
			"G#", "A", "A#", "B" };

static Eventp alloc_event(void) {
  Eventp e = reuse;

  if (e == NULL)
    e = (Eventp)malloc(sizeof(Event));
  else
    reuse = e->next;
  e->next = NULL;
  e->prev = NULL;
  return e;
}

static void free_event(Eventp e) {
  e->next = reuse;
  reuse = e;
}

void mtm_error(char *s) {
  fprintf(stderr,"Error: %s\n",s);
}

void mtm_header(int fmt, int ntrks, int ldivision) {
  int f;

  maxtrk = ntrks;		/* set number of tracks 		*/
  initTracks();			/* allocate tracks and associated bits 	*/
  format = fmt;
  division = ldivision;
  startTime = (long)(pad * division);	/* note real start time */
  if (ffactor == 0) {
    // no quantum specified, find a reasonable value
    // for this division. Note: some synths produce
    // really grotty values, just divide by 2 until
    // we get an odd number, then double that
    ffactor = division / 2;
    while ((ffactor & 2) == 0)
      ffactor /= 2;
  }
  else if (ffactor < 2)
    ffactor = division / 2;
  else {
    f = division / ffactor;		/* derive quantum in ticks */
    if (f * ffactor != division)
      while (ffactor > 2) {
	ffactor = ffactor >> 1;
	f = division / ffactor;
	if (f * ffactor == division) {
	  break;
	}
      }
    if (f * ffactor == division)
      ffactor = f;
    else
      ffactor = division / 2;
  }
  shortest = ffactor;	// division;
  //  while ((shortest & 1) == 0)
  //    shortest = shortest >> 1;
  /* shortest is now smallest possible note length */
  /*
    printf("Header format=%d ntrks=%d division=%d\n",format,ntrks,division);
    */
}

void mtm_trackstart(void) {
  /*
    printf("Track start\n");
    */
  ct = tracks[trackno];
  ct->process = 0;
  trackno++;
  if (trackno > maxtrk) {
    fprintf(stderr, "mtm: fatal: number of tracks exceeds limit\n");
    fprintf(stderr, "mtm: try using the -t option (or larger -t option)\n");
    exit(1);
  }
  Mf_currtime = startTime;	// allow for padding 
}

long adjust(long t) {
  if (ffactor == 0)
    return t;
  t -= shift;
  return ((t + (ffactor >> 1) - 1) / ffactor) * ffactor;
}

// this routine only provided for debugging, to be invoked by patches
// or manually from gdb
void checkCount(Trackp t) {
  int c=0;
  Eventp e = t->notes->e;
  while (e != NULL) {
    if (e->type == NOTE)
      c++;
    e = e->next;
  }
  if (c != t->notes->count) {
    printf("consistency check failure\n");
  }
}

static void add_to_trk(Eventlistp elp, Eventp e) {
  if (elp->p == NULL)
    elp->e = e;
  else
    elp->p->next = e;
  e->prev = elp->p;
  elp->p = e;
  elp->count++;
}

static void discard(Eventlistp elp, Eventp e) {
  // remove event from track (probably because length=0
  if (e->prev == NULL)
    elp->e = e->next;
  else
    e->prev->next = e->next;
  if (e->next != NULL)
    e->next->prev = e->prev;
  if (elp->p == e)
    elp->p = e->prev;
  free_event(e);
  elp->count--;
  discards++;
}

static void finishTrack(Trackp ct) {
  long t = adjust(Mf_currtime);
  Eventp e = alloc_event();

  e->start = t + division;
  e->remains = e->start;
  e->next = NULL;
  e->type = END;
  add_to_trk(ct->notes, e);
  
  e = alloc_event();
  e->start = t + division;
  e->next = NULL;
  e->type = END;
  add_to_trk(ct->mevnts, e);
}

void mtm_trackend(void) {
  /* append dummy entry to mark end-of-track */
  int i;

  if (format == 0) {
    for (i=0; i<trackno; i++)
      finishTrack(tracks[i]);
  }
  else
    finishTrack(ct);
}

static void prtime(void) {
  int t = Mf_currtime / division, f = Mf_currtime % division;
  
  printf("off: 0x%lx, Time=%d.%d \n", bytecount, t, f);
}

static int getBar(long t) {
  // determine bar number for a given time
  long tt = 0, sect, q, bl;
  int b, m=0;
  Tsigp tp = tsigs;

  q = division;
  bl = 4 * q;		// assume 4 beats/bar initially
  while (tp != NULL) {
    if (tp->when > t)
      break;
    sect = (tp->when - tt) / bl;
    tt += sect * bl;	// start of new section
    m += sect;		// total no of bars
    q = (division * 4) / tp->dnom;
    b = q * tp->num;
    if (b > 0)
      bl = b;
    tp = tp->next;
  }
  sect = (t - tt) / bl;	// length of final section
  tt += sect * bl;
  m += sect;
  return m;
}

static long convertBar(int bar) {
  // given a bar number, find time from start
  int b, bl, q, m=0;
  long tt=0, sect;
  Tsigp tp = tsigs;

  q = division;
  bl = 4 * q;			// initial guess
  while (tp != NULL) {
    sect = (tp->when - tt) / bl;
    if (m + sect >= bar)
      break;
    tt += sect * bl;
    m += sect;
    q = (4 * division) / tp->dnom;
    b = tp->num * q;
    if (b > 0)
      bl = b;
    tp = tp->next;
  }
  tt += (bar - m) * bl;
  return tt;
}
      
void mtm_noteon(int chan, int pitch, int vol) {
  long t = Mf_currtime, len, sb, eb;
  Eventp e, e2;
  
  if (debug) {
    prtime();
    printf("Note on, chan=%d pitch=%d vol=%d\n",chan+1,pitch,vol);
  }
  if (format == 0) {
    ct = tracks[chan+1];	/* split notes into tracks based on channel */
    if (trackno <= chan+1)
      trackno = chan+2;
  }
  if (vol == 0) {		/* note off message */
    if (ct->notes->p == NULL) {
      if (warnFlag) {
	prtime();
	printf("end note before first start note\n");
      }
      return;	// ignore end note event
    }
    for (e=ct->notes->p; e; e=e->prev)
      if (e->type == NOTE && e->pitch == pitch && e->end < 0)
	break;
    if (e == NULL) {
      int octave = pitch / 12;
      int note = pitch - 12 * octave;
  
      if (warnFlag) {
	prtime();
	printf("can't find matching start note, trk=%d, pitch=%s%d\n",
	       chan, nnam[note], octave);
      }
      return;			/* discard unmatched note */
    }
    while (e->start >= t)	/* extend zero-length notes to minimum */
      t += ffactor;
    if (stretch > 0) {
      len = t - e->start;
      len += (len * stretch) / 100;
      t = e->start + len;
    }
    e->end = adjust(t);
    if (piece_end < e->end)
      piece_end = e->end;
    sb = getBar(e->start);
    eb = getBar(e->end-1);
    while (sb < eb) {
      // note spans one or more bar lines
      sb++;
      t = convertBar(sb);	/* find end of starting bar */
      e2 = alloc_event();
      *e2 = *e;			/* copy details across */
      e->end = t;		/* make first one end here */
      e->flags = F_TIEF;
      e2->start = t;
      e2->remains = t;
      e2->flags = F_TIEB;	/* note backward tie */
      add_to_trk(ct->notes, e2);
      ct->process = 1;
      e = e2;
    }
    if (e->end <= e->start) {
      // zero or negative length note ??
      if (zflag)
	discard(ct->notes, e);
      else
	e->end = e->start + ffactor;
    }
  }
  else {
    e = alloc_event();
    e->start = adjust(t);
    e->remains = e->start;	/* note start time */
    e->vel = vol;
    e->pitch = pitch;
    e->type = NOTE;
    e->end = -1;		/* unterminated */
    e->flags = 0;
    add_to_trk(ct->notes, e);
    ct->process = 1;
    if (pitch < ct->minp)
      ct->minp = pitch;
    if (pitch > ct->maxp)
      ct->maxp = pitch;
  }
}

void mtm_noteoff(int chan, int pitch, int vol) {
  mtm_noteon(chan, pitch, 0);	/* map all real note-offs to noteon with vel=0 */
}

void mtm_pressure(int chan, int pitch, int press) {
  /*
    prtime();
    printf("Pressure, chan=%d pitch=%d press=%d\n",chan+1,pitch,press);
    */
}

void mtm_program(int chan, int program) {
  /*
    prtime();
    printf("Program, chan=%d program=%d\n",chan+1,program);
    */
  long t = adjust(Mf_currtime);
  Eventp e = alloc_event();

  e->start = t;
  e->type = PATCH;
  e->pitch = program;
  add_to_trk(ct->mevnts, e);
}

void mtm_parameter(int chan, int control, int value) {
  /*
    prtime();
    printf("Parameter, chan=%d c1=%d c2=%d\n",chan+1,control,value);
    */
  long t = adjust(Mf_currtime);
  Eventp e = alloc_event();
  e->start = t;
  e->type = PARAM;
  e->pitch = control;
  e->vel = value;
  add_to_trk(ct->mevnts, e);
}

void mtm_pitchbend(int chan, int msb, int lsb) {
  /*
    prtime();
    printf("Pitchbend, chan=%d msb=%d lsb=%d\n",chan+1,msb,lsb);
    */
  long t = adjust(Mf_currtime);
  Eventp e = alloc_event();
  e->start = t;
  e->type = BEND;
  e->pitch = msb;
  e->vel = lsb;
  add_to_trk(ct->mevnts, e);
}

void mtm_chanpressure(int chan, int press) {
  long t = adjust(Mf_currtime);
  Eventp e = alloc_event();

  e->start = t;
  e->type = PITCH;
  e->pitch = press;
  add_to_trk(ct->mevnts, e);
  /*
    prtime();
    printf("Channel pressure, chan=%d pressure=%d\n",chan+1,press);
    */
}

void mtm_sysex(int leng, char *mess) {
  /*
    prtime();
    printf("Sysex, leng=%d\n",leng);
    */
}

void mtm_metamisc(int type, int leng, char *mess) {
  /*
    prtime();
    printf("Meta event, unrecognized, type=0x%02x leng=%d\n",type,leng);
    */
}

void mtm_seqspecific(int type, int leng, char *mess) {
  /*
    prtime();
    printf("Meta event, sequencer-specific, type=0x%02x leng=%d\n",type,leng);
    */
}

void mtm_text(int type, int leng, char *mess) {
  static char *ttype[] = {
    NULL,
    "Text Event",		/* type=0x01 */
    "Copyright Notice",		/* type=0x02 */
    "Sequence/Track Name",	/* type=0x03 */
    "Instrument Name",		/* type=0x04 */
    "Lyric",			/* type=0x05 */
    "Marker",			/* type=0x06 */
    "Cue Point",		/* type=0x07 */
    "Unrecognized"
  };
  int unrecognized = (sizeof(ttype)/sizeof(char *)) - 1;
  long t = adjust(Mf_currtime);
  Eventp e = alloc_event();

  if ( type < 1 || type > unrecognized )
    type = unrecognized;

  e->start = t;
  e->type = TEXT;
  e->pitch = type;
  mess[leng] = 0;		/* just in case! */
  e->text = strdup(mess);
  add_to_trk(ct->mevnts, e);
}

void mtm_seqnum(int num) {
  /*
    prtime();
    printf("Meta event, sequence number = %d\n",num);
    */
}

void mtm_eot(void) {
  /*
    prtime();
    printf("Meta event, end of track\n");
    */
}

void mtm_keysig(int sf, int mi) {
  if (Mf_currtime == startTime)
    addKey(0, (sf & 0x80) ? (sf | ~0xff) : sf);
  else
    addKey(Mf_currtime, (sf & 0x80) ? (sf | ~0xff) : sf);

}

void mtm_tempo(long ltempo) {
  Eventp e = alloc_event();
  long t = adjust(Mf_currtime);

  e->type = TEMPO;
  e->pitch = 60000000 / ltempo;
  e->start = t;
  add_to_trk(ct->mevnts, e);
}

void mtm_timesig(int nn, int dd, int cc, int bb) {
  int denom = 1;

  if (ct != tracks[0])		// ignore timesigs on other than trk 0
    return;
  while ( dd-- > 0 )
    denom *= 2;
  if (tsigs == NULL)
    addTsig(0, nn, denom);
  else
    addTsig(Mf_currtime, nn, denom);
}

void mtm_smpte(int hr, int mn, int se, int fr, int ff) {
  /*
    prtime();
    printf("SMPTE, hour=%d minute=%d second=%d frame=%d fract-frame=%d\n",
           hr,mn,se,fr,ff);
   */
}

void mtm_arbitrary(int leng, char *mess)
{
  /*
    prtime();
    printf("Arbitrary bytes, leng=%d\n",leng);
    */
}

/* read and return one character from MIDI file. Return EOF on end of file
 * or read error. */
int mtm_getc(void) {
  int ch = fgetc(inf);
  bytecount++;
  debugbuff[bytecount & 15] = ch;
  return ch;
}

// this routine provided for debugging, invoked manually from gdb.
void dumpTrack(Eventp ep) {
  while (ep != NULL) {
    printf("event 0x%08lx: remains=%ld, type=%d, start=%ld, end=%ld pitch=%d\n",
	   (long)ep, ep->remains, ep->type, ep->start, ep->end, ep->pitch);
    ep = ep->next;
  }
}

FILE *efopen(char *name) {
  FILE *f;
  const char *errmess;

  strcpy(ifile, name);
  if ((f=fopen(ifile, "rb")) == NULL) {
    sprintf(ifile, "%s.mid", name);
    if ((f = fopen(ifile, "rb")) == NULL) {
      errmess = strerror(errno);
      (void) fprintf(stderr,"Cannot open '%s'!\n", name);
      (void) fprintf(stderr,"Reason: %s\n",errmess);
      exit(1);
    }
  }
  return(f);
}
